/*
 * Segment 2 (KEx),  SWE 2.Statischer Import Netz
 * Copyright (C) 2007 BitCtrl Systems GmbH 
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weienfelser Strae 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.kex.bwsin.konfig;

import java.util.ArrayList;
import java.util.List;

import org.geotools.feature.Feature;

import de.bsvrz.kex.bwsin.konfig.okstra.Okstra;
import de.bsvrz.kex.bwsin.konfig.okstra.OkstraFeatures;
import de.bsvrz.kex.bwsin.konfig.okstra.OkstraNetzknoten;
import de.bsvrz.kex.bwsin.konfig.okstra.OkstraNullpunkt;
import de.bsvrz.kex.bwsin.konfig.okstra.OkstraStrassenElement;
import de.bsvrz.kex.bwsin.konfig.okstra.OkstraVerbindungspunkt;
import de.bsvrz.kex.bwsin.konfig.okstra.OkstraVerkehrsRichtung;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Internes Stra&szlig;enelement.
 * 
 * @author BitCtrl Systems GmbH, Gieseler
 * @version $Id: StrassenElement.java 10402 2008-07-10 09:07:00Z gieseler $
 * 
 */
public class StrassenElement implements Cloneable {

	/**
	 * DebugLogger fr Debug-Ausgaben.
	 */
	private static final Debug DEBUG = Debug.getLogger();

	/**
	 * Die in der OKSTRA-Datenbasis benutzten Trennzeichen bei mehreren
	 * Locationen.
	 */
	private static final String TRENNZEICHEN_LOCATIONEN = "[,;]"; //$NON-NLS-1$

	/**
	 * Die Feature-ID des Start-Verbindungspunktes.
	 */
	protected String startVP = ""; //$NON-NLS-1$

	/**
	 * Die Feature-ID des End-Verbindungspunktes.
	 */
	protected String endeVP = ""; //$NON-NLS-1$

	/**
	 * Das assoziierte Feature der Okstra-Databasis.
	 */
	private Feature feature;

	/**
	 * Liste von Stra&szlig;enabschnitten, in denen das Stra&szlig;enelement
	 * enthalten ist.
	 */
	private final List<StrassenAbschnitt> inAbschnitten = new ArrayList<StrassenAbschnitt>();

	/**
	 * Die Verkehrsrichtung in Bezug zur Stationierungsrichtung.
	 */
	private Verkehrsrichtung richtung = Verkehrsrichtung.RICHTUNG_UNBEKANNT;

	/**
	 * Flag, ob das Stra&szlig;enelement intern erzeugt wurde.
	 */
	private boolean internErzeugt = false;

	/**
	 * Location, auf die sich das Stra&szlig;enelement zubewegt.
	 */
	private Location nachLocation = null;

	/**
	 * Privater Konstruktor.
	 */
	private StrassenElement() {
		/**
		 * 
		 */
	}

	/**
	 * Erzeugt ein Strassenelement aus einem Feature vom Typ
	 * 'vrz3:StrassenelementInASS'.
	 * 
	 * @param feature
	 *            Strassenelement-Feature
	 * @throws ImportException
	 *             bei Ausnahmen
	 */
	public StrassenElement(Feature feature) throws ImportException {
		this.feature = feature;
		startVP = Okstra.getRefFeatureId((String) feature
				.getAttribute(OkstraStrassenElement.VON_VP));
		endeVP = Okstra.getRefFeatureId((String) feature
				.getAttribute(OkstraStrassenElement.NACH_VP));
		bestimmeRichtung();
	}

	/**
	 * Bestimmt die Netzknoten zwischen denen ein Stra&szlig;enelement
	 * verl&auml;uft.
	 * 
	 * Dazu wird folgende Relation aufgel&ouml;st: <br>
	 * Netzknoten1:
	 * StrassenelementInASS.beginnt_bei_VP.in_Nullpunkt.in_Netzknoten.Bezeichnung_kurz
	 * <br>
	 * Netzknoten2:
	 * StrassenelementInASS.endet_bei_VP.in_Nullpunkt.in_Netzknoten.Bezeichnung_kurz
	 * 
	 * @return Liste von Feature-ID's der Netzknoten
	 */
	public List<String> bestimmeNetzknoten() {
		List<String> nliste = new ArrayList<String>();

		String nk = bestimmeNetzknotenVerbindungspunkt(startVP);
		if (nk != null) {
			nliste.add(nk);
		}

		nk = bestimmeNetzknotenVerbindungspunkt(endeVP);
		if (nk != null) {
			nliste.add(nk);
		}

		return nliste;
	}

	/**
	 * Bestimmt den Netzknoten zu einem Verbindungspunkt.
	 * 
	 * Dazu wird folgende Relation aufgel&ouml;st:
	 * Verbindungspunkt.in_Nullpunkt.in_Netzknoten.Bezeichnung_kurz
	 * 
	 * @param vpid
	 *            Feature-ID des verbindungspunktes
	 * @return Feature-ID des Netzknotens
	 */
	private String bestimmeNetzknotenVerbindungspunkt(String vpid) {
		Feature vpf;
		try {
			vpf = Okstra.getInstance().getFeature(
					OkstraFeatures.VERBINDUNGSPUNKTE, vpid);

			if (vpf == null) {
				DEBUG.error("Der Verbindungspunkt " + vpid + //$NON-NLS-1$
						" existiert nicht"); //$NON-NLS-1$
				return null;
			}

			if (vpf.getAttribute(OkstraVerbindungspunkt.NULLPUNKT) == null) {
				DEBUG.error("Der Verbindungspunkt " + vpid + //$NON-NLS-1$
						" besitzt kein Attribut '"
						+ OkstraVerbindungspunkt.NULLPUNKT + "'"); //$NON-NLS-1$
				return null;
			}

			Feature nullpunkt = Okstra.getFeatureReferenz(vpf
					.getAttribute(OkstraVerbindungspunkt.NULLPUNKT));

			if (nullpunkt == null) {
				DEBUG
						.error("Das referenzierte Feature " + OkstraVerbindungspunkt.NULLPUNKT + //$NON-NLS-1$
								" existiert nicht fr " + vpf.getID()); //$NON-NLS-1$
				return null;
			}

			Feature netzknoten = Okstra.getFeatureReferenz(nullpunkt
					.getAttribute(OkstraNullpunkt.NETZKNOTEN));

			return (String) netzknoten.getAttribute(OkstraNetzknoten.NUMMER);
		} catch (ImportException e) {
			DEBUG.error(e.getMessage());
			return null;
		}
	}

	/**
	 * Bestimmt die Verkehrsrichtung des Stra&szlig;enelementes aus den
	 * Feature-Attributen.
	 * 
	 * @return Verkehrsrichtung
	 * 
	 * @throws ImportException
	 *             wenn die Richtung nicht bestimmt werden kann
	 */
	public Verkehrsrichtung bestimmeRichtung() throws ImportException {
		String r = Okstra.getInstance().getSpecialAttribute2(
				OkstraFeatures.STRASSENELEMENT, feature.getID(),
				OkstraStrassenElement.RICHTUNG);

		if (r == null) {
			throw new ImportException("Die Richtung des Straenelementes " + //$NON-NLS-1$
					feature.getID() + " kann nicht bestimmt werden (null)"); //$NON-NLS-1$
		}

		if (r.equalsIgnoreCase(OkstraVerkehrsRichtung.RICHTUNG_IN_STATRICHTUNG)) {
			richtung = Verkehrsrichtung.RICHTUNG_IN_STATIONIERUNGSRICHTUNG;
		} else if (r
				.equalsIgnoreCase(OkstraVerkehrsRichtung.RICHTUNG_BEIDE_RICHTUNGEN)) {
			richtung = Verkehrsrichtung.RICHTUNG_BEIDE_RICHTUNGEN;
		} else if (r
				.equalsIgnoreCase(OkstraVerkehrsRichtung.RICHTUNG_GEGEN_STATRICHTUNG)) {
			richtung = Verkehrsrichtung.RICHTUNG_GEGEN_STATIONIERUNGSRICHTUNG;
		} else {
			throw new ImportException("Die Richtung des Straenelementes " + //$NON-NLS-1$
					feature.getID() + " kann nicht bestimmt werden"); //$NON-NLS-1$
		}

		return richtung;
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @see java.lang.Object#clone()
	 */
	@Override
	public StrassenElement clone() {
		DEBUG.finest("Das Strassenelement '" + this + "' wird geclont"); //$NON-NLS-1$//$NON-NLS-2$
		StrassenElement neu = new StrassenElement();
		neu.feature = this.feature;
		neu.startVP = this.startVP;
		neu.endeVP = this.endeVP;
		neu.richtung = this.richtung;
		return neu;
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		final StrassenElement other = (StrassenElement) obj;
		if (endeVP == null) {
			if (other.endeVP != null) {
				return false;
			}
		} else if (!endeVP.equals(other.endeVP)) {
			return false;
		}
		if (feature == null) {
			if (other.feature != null) {
				return false;
			}
		} else if (!feature.getID().equals(other.feature.getID())) {
			return false;
		}

		if (nachLocation == null) {
			if (other.nachLocation != null) {
				return false;
			}
		} else if (!nachLocation.equals(other.nachLocation)) {
			return false;
		}

		if (richtung == null) {
			if (other.richtung != null) {
				return false;
			}
		} else if (!richtung.equals(other.richtung)) {
			return false;
		}
		if (startVP == null) {
			if (other.startVP != null) {
				return false;
			}
		} else if (!startVP.equals(other.startVP)) {
			return false;
		}
		return true;
	}

	/**
	 * Erzeugt eine Kopie des Stra&szlig;enelementes f&uuml;r die Gegenrichtung.
	 * Die Richtung dieses Stra&szlig;enelementes wird auf
	 * IN_STATIONIERUNGSRICHTUNG gesetzt. Das neue Stra&szlig;enelementes
	 * erh&auml;lt die Verkehrsrichtung GEGEN_STATIONIERUNGSRICHTUNG. Fr das
	 * neue Stra&szlig;enelement werden die Verbindungspunkte entsprechend
	 * angepa&szlig;t.
	 * 
	 * @return neues {@link StrassenElement}
	 * @throws ImportException
	 *             wenn kein Stra&szlig;enelement angelegt werden kann
	 */
	public StrassenElement erzeugeGegenrichtung() throws ImportException {
		StrassenElement neuesElement = new StrassenElement(this.getFeature());

		// Verkehrsrichtung anpassen
		this.richtung = Verkehrsrichtung.RICHTUNG_IN_STATIONIERUNGSRICHTUNG;
		neuesElement.richtung = Verkehrsrichtung.RICHTUNG_GEGEN_STATIONIERUNGSRICHTUNG;

		// Verbindungspunkte anpassen
		neuesElement.vpUmkehren();

		neuesElement.internErzeugt = true;

		return neuesElement;
	}

	/**
	 * Gibt den Verbindungspunkt zur&uuml;ck, an dem das Stra&szlig;ensegment
	 * endet.
	 * 
	 * @return Feature-Id des Verbindungspunktes
	 */
	public String getEndeVP() {
		return endeVP;
	}

	/**
	 * Gibt das mit diesem Stra&szlig;enelement assoziierte Feature zur&uuml;ck.
	 * 
	 * @return Feature
	 */
	public Feature getFeature() {
		return feature;
	}

	/**
	 * Bestimmt die Locationen, auf die das Stra&szlig;element zuf&auml;hrt aus
	 * den Feature-Attributen.
	 * 
	 * @return Liste von Locationen
	 * @throws ImportException
	 *             bei Ausnahmen
	 */
	public List<Location> getFeatureNachLocationen() throws ImportException {
		String[] locs = null;
		List<Location> locations = new ArrayList<Location>();

		if (!istASSFeature()) {
			return locations;
		}

		if (richtung == Verkehrsrichtung.RICHTUNG_IN_STATIONIERUNGSRICHTUNG) {
			String loc = (String) feature
					.getAttribute(OkstraStrassenElement.TMCLOC_RICHTUNG);
			if (loc == null || loc.length() == 0) {
				throw new ImportException("Das Straenelement " + //$NON-NLS-1$
						feature.getID()
						+ " besitzt ein inkonsistentes Location-Attribut."); //$NON-NLS-1$
			}

			locs = loc.split(TRENNZEICHEN_LOCATIONEN);
		}

		if (richtung == Verkehrsrichtung.RICHTUNG_GEGEN_STATIONIERUNGSRICHTUNG) {
			String loc = (String) feature
					.getAttribute(OkstraStrassenElement.TMCLOC_GEGENRICHTUNG);
			if (loc == null || loc.length() == 0) {
				throw new ImportException("Das Straenelement " + //$NON-NLS-1$
						feature.getID()
						+ " besitzt ein inkonsistentes Location-Attribut."); //$NON-NLS-1$
			}

			locs = loc.split(TRENNZEICHEN_LOCATIONEN);
		}

		if (locs != null && locs.length > 0) {
			for (int i = 0; i < locs.length; i++) {
				try {
					locations.add(new Location(locs[i].trim()));
				} catch (Exception e) {
					throw new ImportException(e.getMessage());
				}
			}

			return locations;
		}

		throw new ImportException(
				"Die Location-Richtung des Straenelementes " + //$NON-NLS-1$
						feature.getID()
						+ " kann nicht bestimmt werden oder ist ungltig"); //$NON-NLS-1$
	}

	/**
	 * Gibt die Feature-ID des mit dem Strassenelement assoziierten Features der
	 * Okstra-Databasis zur&uuml;ck.
	 * 
	 * @return Feature-ID
	 */
	public String getID() {
		return (feature.getID());
	}

	/**
	 * Gibt die Location zur&uuml;ck, auf die das Stra&szlig;enelement
	 * zuf&auml;hrt.
	 * 
	 * @return Locationen, auf die das Stra&szlig;enelement
	 */
	public Location getNachLocation() {
		return nachLocation;
	}

	/**
	 * Gibt die Verkehrsrichtung des Strassenelementes zur&uuml;ck.
	 * 
	 * @return {@link Verkehrsrichtung}
	 */
	public Verkehrsrichtung getRichtung() {
		return richtung;
	}

	/**
	 * Gibt den Verbindungspunkt zur&uuml;ck, an dem das Stra&szlig;ensegment
	 * beginnt.
	 * 
	 * @return Feature-Id des Verbindungspunktes
	 */
	public String getStartVP() {
		return startVP;
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((endeVP == null) ? 0 : endeVP.hashCode());
		result = prime * result
				+ ((feature == null) ? 0 : feature.getID().hashCode());
		result = prime * result + ((richtung == null) ? 0 : richtung.ordinal());
		result = prime * result + ((startVP == null) ? 0 : startVP.hashCode());
		result = prime * result
				+ ((nachLocation == null) ? 0 : nachLocation.hashCode());
		return result;
	}

	/**
	 * Test, ob das Straenelement zu einem &auml;ueren Straensegment
	 * geh&ouml;rt.
	 * 
	 * @return true, wenn das Straenelement zu einem &auml;ueren
	 *         Straensegment geh&ouml;rt, ansonsten false.
	 */
	public boolean istASS() {
		return nachLocation != null;
	}

	/**
	 * Test, ob das Straenelement zu einem &auml;ueren Straensegment
	 * geh&ouml;rt.
	 * 
	 * @return true, wenn das Straenelement zu einem &auml;ueren
	 *         Straensegment geh&ouml;rt, ansonsten false.
	 */
	public boolean istASSFeature() {
		String tmcr = (String) feature
				.getAttribute(OkstraStrassenElement.TMCLOC_RICHTUNG);
		String tmcg = (String) feature
				.getAttribute(OkstraStrassenElement.TMCLOC_GEGENRICHTUNG);

		return (tmcr != null) || (tmcg != null);
	}

	/**
	 * Gibt das Flag zur&uuml;ck, ob es das Stra&szlig;enelement intern erzeugt
	 * wurde.
	 * 
	 * @return true oder false
	 */
	public boolean istInternErzeugt() {
		return internErzeugt;
	}

	/**
	 * Test, ob das Strassenelement zur Bildug von Stra&szlig;enabschnitten
	 * erwendet werden kann. Strassenelemente, die in beiden Fahrtrichtungen
	 * verlaufen, k&ouml;nnen 2 Abschnitten zugeordnet werden.
	 * 
	 * @return Wahr, wenn das Strassenelement zu einem Strassenabschnitt
	 *         geh&ouml;rt.
	 */
	public boolean kannBenutztWerden() {
		if (richtung == Verkehrsrichtung.RICHTUNG_BEIDE_RICHTUNGEN) {
			return inAbschnitten.size() < 2;
		}

		return inAbschnitten.size() == 0;
	}

	/**
	 * Test, ob ein Strassenelement importiert werden kann. Bedingung ist ein
	 * existierender Locationcode zum Netzknoten, in dem dieses (innere)
	 * Stra&szlig;enelement verl&auml;ft.
	 * 
	 * F&uuml;r den ermittelten Netzknoten wird ein Locationcode in der LCL
	 * gesucht, der in der Spalte NETZKNOTEN_NR den Netzknoten enth&auml;lt.
	 * 
	 * @return true, wenn das Stra&szlig;enelement importiert werden kann,
	 *         ansonsten false
	 */
	public boolean kannImportiertWerden() {
		if (istASS()) {
			return true;
		}

		return true;
	}

	/**
	 * Setzt die Location, auf die das Stra&szlig;enelement zuf&auml;hrt.
	 * 
	 * @param location
	 *            Location
	 */
	public void setNachLocation(Location location) {
		nachLocation = location;
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return (feature.getID() + vrString());
	}

	/**
	 * Test, ob sich ein anderes Stra&szlig;enelement auf die gleiche Location
	 * zubewegt.
	 * 
	 * @param element
	 *            anderes Stra&szlig;enelement
	 * @return true, wenn sich das andere Stra&szlig;enelement auf die gleiche
	 *         Location zubewegt, sonst false
	 */
	public boolean ueberpruefeLocationen(final StrassenElement element) {
		if (nachLocation == null) {
			return false;
		}

		return nachLocation.equals(element.getNachLocation());
	}

	/**
	 * Tauscht die Verbindungspunkte der Stra&szlig;enelementes.
	 */
	public void vpUmkehren() {
		// Verbindungspunkte anpassen
		String vpp = startVP;
		startVP = endeVP;
		endeVP = vpp;
	}

	/**
	 * Gibt eine textuelle Repr&auml;sentation der Verkehrsrichtung zur&uuml;ck.
	 * 
	 * @return Verkehrsrichtung als Text
	 */
	private String vrString() {
		if (richtung == Verkehrsrichtung.RICHTUNG_IN_STATIONIERUNGSRICHTUNG) {
			return ("(R)");
		}
		if (richtung == Verkehrsrichtung.RICHTUNG_GEGEN_STATIONIERUNGSRICHTUNG) {
			return ("(G)");
		}
		if (richtung == Verkehrsrichtung.RICHTUNG_BEIDE_RICHTUNGEN) {
			return ("(B)");
		}
		if (richtung == Verkehrsrichtung.RICHTUNG_UNBEKANNT) {
			return ("(U)");
		}
		return ("?");
	}

	/**
	 * F&uuml;gt dem Stra&szlig;enelement einen Verweis auf einen Stra&szlig;en-
	 * abschnitt hinzu, zu dem es geh&ouml;hrt.
	 * 
	 * @param abschnitt
	 *            {@link StrassenAbschnitt}
	 */
	public void zuAbschnitt(StrassenAbschnitt abschnitt) {
		inAbschnitten.add(abschnitt);
	}
}
