/*
 * Segment 2 (KEx),  SWE 2.Statischer Import Netz
 * Copyright (C) 2007 BitCtrl Systems GmbH 
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weienfelser Strae 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.kex.bwsin.konfig.kb;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.GregorianCalendar;

import org.geotools.feature.Feature;

import de.bsvrz.kex.bwsin.konfig.ImportException;
import de.bsvrz.kex.bwsin.konfig.okstra.IllegalAttributeConversionException;
import de.bsvrz.kex.bwsin.konfig.okstra.Okstra;
import de.bsvrz.kex.bwsin.konfig.okstra.OkstraDauer;
import de.bsvrz.kex.bwsin.konfig.okstra.OkstraEinfacherZeitraum;
import de.bsvrz.kex.bwsin.konfig.okstra.OkstraFeatures;
import de.bsvrz.kex.bwsin.konfig.okstra.OkstraStartDatum;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Zeitbereich Verkehrsbeschr&auml;nkungen.
 * 
 * Die Spezifikation im DaV erfolgt durch die Angabe von 5 Terminfeldern, wobei
 * im ersten Feld die Minute [0;59], im zweiten Feld die Stunde [0;23], im
 * dritten Feld der Tag [1;31], im viertem Feld der Monat [1;12] und im letzten
 * Feld der Wochentag {Mo,Di,Mi,Do,Fr,Sa,So} angegeben werden kann. Pro
 * Terminfeld sind mehrere Eintr&auml;ge, die durch Kommata getrennt werden
 * m&uuml;ssen, m&ouml;glich. Die Eingabe eines '*' deckt den gesamten Bereich
 * eines Terminfeldes ab. Mit dem Zeichen '-' k&ouml;nnen Teilbereiche abgedeckt
 * werden. Die Angaben aller Terminfelder sind UND-verknpft.
 * <p>
 * Beispiele:
 * <p>
 * Die Angabe '0 2 * * *' spezifiziert, dass die Aktion t&auml;glich um 02:00
 * Uhr durchgef&uuml;hrt werden soll.
 * <p>
 * Die Angabe '30 2 * * Mo-Mi,Fr' bewirkt, dass die Aktion montags, dienstags,
 * mittwochs und freitags um 02:30 durchgef&uuml;hrt werden soll.
 * <p>
 * Die Angabe '0 3 1,15 * *' spezifiziert, dass die Aktion jeden 1. und 15.
 * eines Monats um 03:00 Uhr durchgef&uuml;hrt werden soll.
 * 
 * @author BitCtrl Systems GmbH, Gieseler
 * @version $Id: Zeitbereich.java 7938 2008-04-03 17:42:19Z gieseler $
 * 
 */

public class Zeitbereich {
	/**
	 * Ein Terminfeld des Zeitbereiches. Pro Terminfeld sind mehrere
	 * Eintr&auml;ge, die durch Kommata getrennt werden m&uuml;ssen,
	 * m&ouml;glich. Die Eingabe eines '*' deckt den gesamten Bereich eines
	 * Terminfeldes ab. Mit dem Zeichen '-' k&ouml;nnen Teilbereiche abgedeckt
	 * werden.
	 */
	class ZeitbereichTerminFeld extends
			ArrayList<ZeitbereichTerminFeldEinzelAngabe> {
		/**
		 * serialVersionUID.
		 */
		private static final long serialVersionUID = 1L;

		/**
		 * f&uuml;r die Berechnung zu benutzendes Kalenderfeld.
		 */
		protected final int kalenderFeld;

		/**
		 * Konstruktor.
		 * 
		 * @param kalenderFeld
		 *            f&uuml;r die Berechnung zu benutzendes Kalenderfeld
		 */
		public ZeitbereichTerminFeld(int kalenderFeld) {
			super();
			this.kalenderFeld = kalenderFeld;
		}

		/**
		 * Setzt Beginn und Dauer durch Auswertung der Features.
		 * 
		 * @param featureAnfang
		 *            Das den Beginn repr&auml;sentierende Feature
		 * @param attributAnfang
		 *            Name des Attributes, dass den Anfangswert enth&auml;lt
		 * @param featureDauer
		 *            Das die Dauer repr&auml;sentierende Feature
		 * @param attributDauer
		 *            Name des Attributes, dass den Dauer-Wert enth&auml;lt
		 */
		public void setBeginnUndDauer(final Feature featureAnfang,
				final String attributAnfang, final Feature featureDauer,
				final String attributDauer) {
			try {
				int beginn = Okstra.getIntAttribute(featureAnfang,
						attributAnfang);
				int dauer = Okstra.getIntAttribute(featureDauer, attributDauer);
				setBeginnUndDauer(beginn, dauer);
			} catch (IllegalAttributeConversionException e) {
				Debug.getLogger().error(e.getMessage());
			} catch (IllegalArgumentException e) {
				// Attribut nicht vorhanden -> kein Fehler
			}
		}

		/**
		 * Setzt Beginn und Dauer.
		 * 
		 * @param beginn
		 *            Beginn
		 * @param dauer
		 *            Dauer
		 */
		public void setBeginnUndDauer(int beginn, int dauer) {
			this.add(new ZeitbereichTerminFeldEinzelAngabe(beginn, dauer));
		}

		/**
		 * {@inheritDoc}
		 * 
		 * @see java.lang.Object#toString()
		 */
		@Override
		public String toString() {
			if (this.size() == 0) {
				return "*";
			}

			String feld = "";
			for (int i = 0; i < this.size(); i++) {
				if (i > 0) {
					feld += ",";
				}

				feld += this.get(i).getText(kalenderFeld);
			}

			return feld;
		}
	}

	/**
	 * Eine einzelne Angabe innerhalb eines Terminfeldes.
	 */
	class ZeitbereichTerminFeldEinzelAngabe {

		/** Wert f&uuml;r Beginn. */
		private int beginn = -1;
		/** Wert f&uuml;r Dauer. */
		private int dauer = -1;

		/**
		 * Konstruktor mit Beginn-Wert.
		 * 
		 * @param beginn
		 *            Wert f&uuml;r Beginn
		 */
		protected ZeitbereichTerminFeldEinzelAngabe(final int beginn) {
			this.beginn = beginn;
		}

		/**
		 * Konstruktor mit Beginn und Ende.
		 * 
		 * @param beginn
		 *            Wert f&uuml;r Beginn
		 * @param dauer
		 *            Wert f&uuml;r Dauer
		 */
		protected ZeitbereichTerminFeldEinzelAngabe(final int beginn,
				final int dauer) {
			this.beginn = beginn;
			this.dauer = dauer;
		}

		/**
		 * Gibt den textuellen Wert zur&uuml;ck.
		 * 
		 * @param kalenderFeld
		 *            f&uuml;r die Berechnung zu benutzendes Kalenderfeld
		 * @return texttueller Wert
		 */
		public String getText(final int kalenderFeld) {
			int korrbeginn = beginn;
			// Werte anpassen
			if (kalenderFeld == Calendar.MONTH) {
				korrbeginn = beginn - 1;
			}

			if (dauer != -1) {
				Calendar k = new GregorianCalendar();
				k.set(kalenderFeld, korrbeginn);
				k.add(kalenderFeld, dauer);
				return String.valueOf(korrbeginn) + "-"
						+ String.valueOf(k.get(kalenderFeld));
			}
			return String.valueOf(korrbeginn);
		}

	}

	/**
	 * Erweitert den Zeitbereich um spezielle Methoden f&uuml;r die
	 * Wochentags-Behandlung.
	 */
	protected class ZeitbereichWochentag extends ZeitbereichTerminFeld {

		/**
		 * serialVersionUID.
		 */
		private static final long serialVersionUID = 1L;

		/**
		 * Das OKSTRA-Attribut "Tag_in_der_Woche" gibt den Wochentag in
		 * folgender Codierung an: 1 = Sonntag 2 = Montag 3 = Dienstag 4 =
		 * Mittwoch 5 = Donnerstag 6 = Freitag 7 = Samstag.
		 * 
		 * OKSTRA-Sonntag.
		 */
		protected static final int OKSTRA_SONNTAG = 1;

		/** OKSTRA-Montag. */
		protected static final int OKSTRA_MONTAG = 2;

		/** OKSTRA-Dienstag. */
		protected static final int OKSTRA_DIENSTAG = 3;

		/** OKSTRA-Mittwoch. */
		protected static final int OKSTRA_MITTWOCH = 4;

		/** OKSTRA-Donnerstag. */
		protected static final int OKSTRA_DONNERSTAG = 5;

		/** OKSTRA-Freitag. */
		protected static final int OKSTRA_FREITAG = 6;

		/** OKSTRA-Sonnabend. */
		protected static final int OKSTRA_SONNABEND = 7;

		/** DaV-Sonntag. */
		protected static final String DAV_SONNTAG = "So";

		/** DAV-Montag. */
		protected static final String DAV_MONTAG = "Mo";

		/** DAV-Dienstag. */
		protected static final String DAV_DIENSTAG = "Di";

		/** DAV-Mittwoch. */
		protected static final String DAV_MITTWOCH = "Mi";

		/** DAV-Donnerstag. */
		protected static final String DAV_DONNERSTAG = "Do";

		/** DAV-Freitag. */
		protected static final String DAV_FREITAG = "Fr";

		/** DAV-Sonnabend. */
		protected static final String DAV_SONNABEND = "Sa";

		/**
		 * Konstruktor.
		 */
		public ZeitbereichWochentag() {
			super(Calendar.DAY_OF_WEEK);
		}

		/**
		 * Gibt die Wochentage als Eintrag f&uuml;r einen DaV-Zeitbereich
		 * zur&uuml:ck.
		 * 
		 * @param zeitbereichEinzelAngabe
		 *            einzelne Zeitbereichsangabe
		 * @return Wochentage als Eintrag f&uuml;r einen DaV-Zeitbereich
		 */
		private String getDavWochentage(
				ZeitbereichTerminFeldEinzelAngabe zeitbereichEinzelAngabe) {
			int korrbeginn;
			try {
				korrbeginn = konvertiereOkstraWochentagNachCalendar(zeitbereichEinzelAngabe.beginn);

				if (zeitbereichEinzelAngabe.dauer != -1) {
					Calendar k = new GregorianCalendar();
					k.set(kalenderFeld, korrbeginn);
					k.add(kalenderFeld, zeitbereichEinzelAngabe.dauer);
					return konvertiereCalendarWochentagNachDav(korrbeginn)
							+ "-"
							+ konvertiereCalendarWochentagNachDav(k
									.get(kalenderFeld));
				}

				return konvertiereCalendarWochentagNachDav(korrbeginn);
			} catch (ImportException e) {
				Debug.getLogger().error(e.getMessage());
				return "*";
			}
		}

		/**
		 * Konvertiert den Wochentag des JAVA-Kalenders in den entsprechenden
		 * Wochentag f&uuml;r einen DaV-Zeitbereich.
		 * 
		 * @param calendartag
		 *            Wochentag aus JAVA-Kalender
		 * @return zugeh&ouml;riger Wochentag f&uuml;r den DaV-Zeitbereich
		 * @throws ImportException
		 *             wenn der Wochentag nicht zugeordnet werden kann.
		 */
		private String konvertiereCalendarWochentagNachDav(int calendartag)
				throws ImportException {
			switch (calendartag) {
			case Calendar.SUNDAY:
				return DAV_SONNTAG;
			case Calendar.MONDAY:
				return DAV_MONTAG;
			case Calendar.TUESDAY:
				return DAV_DIENSTAG;
			case Calendar.WEDNESDAY:
				return DAV_MITTWOCH;
			case Calendar.THURSDAY:
				return DAV_DONNERSTAG;
			case Calendar.FRIDAY:
				return DAV_FREITAG;
			case Calendar.SATURDAY:
				return DAV_SONNABEND;
			default:
				throw new ImportException("Der Wochentag " + calendartag
						+ " kann nicht interpretiert werden");
			}
		}

		/**
		 * Konvertiert den Wochentag der OKSTRA-Zeitabgabe in den entsprechenden
		 * Wochentag f&uuml;r den Java-Calender.
		 * 
		 * @param okstrawochentag
		 *            Wochentag aus OKSTRA
		 * @return zugeh&ouml;riger Wochentag Java-Calender
		 * @throws ImportException
		 *             wenn der Wochentag nicht zugeordnet werden kann.
		 */
		private int konvertiereOkstraWochentagNachCalendar(int okstrawochentag)
				throws ImportException {
			switch (okstrawochentag) {
			case OKSTRA_SONNTAG:
				return Calendar.SUNDAY;
			case OKSTRA_MONTAG:
				return Calendar.MONDAY;
			case OKSTRA_DIENSTAG:
				return Calendar.TUESDAY;
			case OKSTRA_MITTWOCH:
				return Calendar.WEDNESDAY;
			case OKSTRA_DONNERSTAG:
				return Calendar.THURSDAY;
			case OKSTRA_FREITAG:
				return Calendar.FRIDAY;
			case OKSTRA_SONNABEND:
				return Calendar.SATURDAY;
			default:
				throw new ImportException("Der OKSTRA-Wochentag "
						+ okstrawochentag + " kann nicht interpretiert werden");
			}
		}

		/**
		 * {@inheritDoc}
		 * 
		 * @see java.lang.Object#toString()
		 */
		@Override
		public String toString() {
			if (this.size() == 0) {
				return "*";
			}

			String feld = "";
			for (int i = 0; i < this.size(); i++) {
				if (i > 0) {
					feld += ",";
				}

				feld += getDavWochentage(this.get(i));
			}

			return feld;
		}
	}

	/**
	 * serialVersionUID.
	 */
	private static final long serialVersionUID = 1L;

	/**
	 * Feld: Minute.
	 */
	ZeitbereichTerminFeld minute = new ZeitbereichTerminFeld(Calendar.MINUTE);

	/**
	 * Feld: Stunde.
	 */
	ZeitbereichTerminFeld stunde = new ZeitbereichTerminFeld(
			Calendar.HOUR_OF_DAY);

	/**
	 * Feld: Tag.
	 */
	ZeitbereichTerminFeld tag = new ZeitbereichTerminFeld(Calendar.DAY_OF_MONTH);

	/**
	 * Feld: Monat.
	 */
	ZeitbereichTerminFeld monat = new ZeitbereichTerminFeld(Calendar.MONTH);

	/**
	 * Feld: Wochentag.
	 */
	ZeitbereichWochentag wochentag = new ZeitbereichWochentag();

	/**
	 * Das zugeh&ouml;rige OKSTRA-Feature.
	 */
	Feature feature = null;

	/**
	 * Standard-Konstruktor. Erzeugt einen immer gltigen Zeitbereich.
	 */
	public Zeitbereich() {
		/** */
	}

	/**
	 * Erzeugt einen Zeitbereiche aus einem Feature. Das Feature muss vom Typ
	 * okstra:einfacher_Zeitraum oder okstra:komplexer_Zeitraum sein.
	 * 
	 * @param feature
	 *            Feature
	 */
	public Zeitbereich(Feature feature) {
		this.feature = feature;
		if (OkstraFeatures.EINFACHER_ZEITRAUM.istFeatureTyp(feature
				.getFeatureType())) {
			try {
				erzeugeAusEinfacherZeitraum();
			} catch (ImportException e) {
				Debug.getLogger().error(e.getMessage());
				Debug.getLogger().warning(
						"Der Zeitbereich wird als immer gltig benutzt");
			}
		} else {
			throw new IllegalArgumentException(
					"Der bergebene Feature-Typ wird nicht untersttzt");
		}
	}

	/**
	 * Erzeugt die Daten des Zeitbereiches aus einem Feature vom Typ
	 * okstra:einfacher_Zeitraum.
	 * 
	 * @throws ImportException
	 *             bei Ausnahmen
	 */
	private void erzeugeAusEinfacherZeitraum() throws ImportException {
		// bestimme Startdatum
		String startdatumattr = (String) feature
				.getAttribute(OkstraEinfacherZeitraum.STARTDATUM);
		if ((startdatumattr == null) || startdatumattr.length() == 0) {
			throw new ImportException("Das Feature " + feature.getID() + //$NON-NLS-1$
					" besitzt kein Startdatum-Attribut"); //$NON-NLS-1$
		}

		Feature startdatum = Okstra.getFeatureReferenz(startdatumattr);
		if (startdatum == null) {
			throw new ImportException(
					"Das referenzierte Feature " + startdatumattr + //$NON-NLS-1$
							" existiert nicht"); //$NON-NLS-1$
		}

		// bestimme Dauer
		String dauerattr = (String) feature
				.getAttribute(OkstraEinfacherZeitraum.DAUER);
		if ((dauerattr == null) || dauerattr.length() == 0) {
			throw new ImportException("Das Feature " + feature.getID() + //$NON-NLS-1$
					" besitzt kein Dauer-Attribut"); //$NON-NLS-1$
		}

		Feature dauer = Okstra.getFeatureReferenz(dauerattr);
		if (dauer == null) {
			throw new ImportException(
					"Das referenzierte Feature " + startdatumattr + //$NON-NLS-1$
							" existiert nicht"); //$NON-NLS-1$
		}

		minute.setBeginnUndDauer(startdatum, OkstraStartDatum.MINUTE, dauer,
				OkstraDauer.MINUTEN);
		stunde.setBeginnUndDauer(startdatum, OkstraStartDatum.STUNDE, dauer,
				OkstraDauer.STUNDEN);
		tag.setBeginnUndDauer(startdatum, OkstraStartDatum.WOCHENTAG, dauer,
				OkstraDauer.TAGE);
		monat.setBeginnUndDauer(startdatum, OkstraStartDatum.MONAT, dauer,
				OkstraDauer.MONATE);
		wochentag.setBeginnUndDauer(startdatum, OkstraStartDatum.WOCHENTAG,
				dauer, OkstraDauer.TAGE);
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return minute + " " + stunde + " " + tag + " " + monat + " "
				+ wochentag;
	}
}
